import math
from enum import Enum
from typing import Union, Tuple, Any

from AoE2ScenarioParser.datasets.buildings import BuildingInfo
from AoE2ScenarioParser.datasets.heroes import HeroInfo
from AoE2ScenarioParser.datasets.other import OtherInfo
from AoE2ScenarioParser.datasets.support.info_dataset_base import InfoDatasetBase
from AoE2ScenarioParser.datasets.units import UnitInfo
from AoE2ScenarioParser.helper import exceptions
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.objects.support.tile import Tile

""" =============================================================
============================ COORDS =============================
=============================================================="""


def xy_to_i(x, y, map_size) -> int:
    if max(x, y) >= map_size or min(x, y) < 0:
        raise ValueError("X and Y need to be: 0 <= n < map_size")
    return x + y * map_size


def i_to_xy(i, map_size) -> Tile:
    if i < 0 or i >= pow(map_size, 2):
        raise ValueError(f"X and Y need to be: 0 <= n ({i}) < map_size ({map_size})")
    return Tile(int(i % map_size), int(i / map_size))


""" =============================================================
============================= OTHER =============================
=============================================================="""


def value_is_valid(value: Any) -> bool:
    """
    Check if value is valid by making sure it's not -1 nor None

    Returns:
        True if value is not None or -1. False if it is either of those.
    """
    return value not in [None, -1]


def values_are_valid(*args: Any) -> bool:
    """
    Check if value is valid by making sure it's not -1 nor None

    Returns:
        True if all values are not None or -1. False if any of them are either of those.
    """
    return any(map(lambda v: v not in [None, -1], args))


def get_enum_from_unit_const(const: int) -> Union[InfoDatasetBase, None]:
    """
    Returns an Enum corresponding with the given Const.

    Arguments:
        const: The constant representing a unit
    """
    enums = [UnitInfo, BuildingInfo, HeroInfo, OtherInfo]
    for enum in enums:
        try:
            return enum.from_id(const)
        except (ValueError, KeyError):
            continue
    return None


def get_int_len(num):
    if num > 0:
        return math.floor(math.log10(num))
    return 0


def exclusive_if(*args):
    """Returns True if exactly one entry is true. False otherwise"""
    return sum(map(bool, args)) == 1


def raise_if_not_int_subclass(values):
    for v in values:
        if not issubclass(v.__class__, int):
            raise TypeError(exceptions.type_error_message(v, issubclass(v.__class__, Enum)))


def validate_coords(x1: int, y1: int, x2: int = None, y2: int = None) -> Tuple[int, int, int, int]:
    """
    Validates given coordinates.

    - Swaps x/y1 with x/y2 if 1 is higher than it's 2 counterpart
    - Sets x/y2 to x/y1 if it's not been set.

    Args:
        x1: The X location of the left corner
        y1: The Y location of the left corner
        x2: The X location of the right corner
        y2: The Y location of the right corner

    Returns:
        The updated values through a tuple as (x1, y1, x2, y2)
    """
    if value_is_valid(x1) and not value_is_valid(x2):
        x2 = x1
    if value_is_valid(y1) and not value_is_valid(y2):
        y2 = y1
    if x1 > x2:
        x1, x2 = x2, x1
        warn("Swapping 'x1' and 'x2' values. Attribute 'x1' cannot be higher than 'x2'")
    if y1 > y2:
        y1, y2 = y2, y1
        warn("Swapping 'y1' and 'y2' values. Attribute 'y1' cannot be higher than 'y2'")
    return x1, y1, x2, y2



from typing import Union, List, Optional, Iterator, Dict
from uuid import UUID

from AoE2ScenarioParser.helper.attr_dict import AttrDict
from AoE2ScenarioParser.helper.helper import values_are_valid, value_is_valid
from AoE2ScenarioParser.objects.data_objects.condition import Condition
from AoE2ScenarioParser.objects.data_objects.effect import Effect
from AoE2ScenarioParser.objects.data_objects.trigger import Trigger
from AoE2ScenarioParser.objects.data_objects.unit import Unit
from AoE2ScenarioParser.objects.support.area import Area
from AoE2ScenarioParser.objects.support.tile import Tile
from AoE2ScenarioParser.scenarios.scenario_store import getters, actions


class DataTriggers:
    _scan_options: Dict[str, List[str]] = {
        "area": ["area", "areas"],
        "tile": ["tile", "tiles"],
        "object": ["object", "objects"],
        "trigger": ["trigger", "triggers"],
    }

    def __init__(self, uuid: UUID) -> None:
        super().__init__()

        self._uuid = uuid

        self.objects = AttrDict()
        self.areas = AttrDict()
        self.tiles = AttrDict()
        self.triggers = AttrDict()

    def discover(self, remove_template_triggers: bool) -> None:
        triggers = getters.get_triggers_by_prefix(
            self._uuid,
            tuple(val for lst in self._scan_options.values() for val in lst)
        )

        for trigger in triggers:
            object_ids = []
            tag = self._resolve_data_trigger_name(trigger)
            for ce in loop_trigger_content(trigger):
                if trigger.name.startswith("area"):
                    if area := self._create_area(ce):
                        self.areas.setdefault(tag, []).append(area)

                elif trigger.name.startswith("tile"):
                    if tiles := self._create_tiles(ce):
                        self.tiles.setdefault(tag, []).extend(tiles)

                elif trigger.name.startswith("object"):
                    object_ids.extend(self._get_unit_ids(ce))
                    if objects := self._get_objects_from_area(ce):
                        self.objects.setdefault(tag, []).extend(objects)
                    if objects := self._get_objects_from_tile(ce):
                        self.objects.setdefault(tag, []).extend(objects)

                elif trigger.name.startswith("trigger"):
                    if found_trigger := self._get_trigger(ce):
                        self.triggers.setdefault(tag, []).append(found_trigger)

            if object_ids:
                self.objects.setdefault(tag, []).extend(getters.get_units(self._uuid, object_ids)[0])

        if remove_template_triggers:
            actions.remove_triggers(self._uuid, [t.trigger_id for t in triggers])

    def _get_objects_from_area(self, ce: Union['Condition', 'Effect']) -> Optional[List[Unit]]:
        if values_are_valid(ce.area_x1, ce.area_y1, ce.area_x2, ce.area_y2):
            return getters.get_units_in_area(self._uuid, ce.area_x1, ce.area_y1, ce.area_x2, ce.area_y2)

    def _get_objects_from_tile(self, ce: Union['Condition', 'Effect']) -> Optional[List[Unit]]:
        if isinstance(ce, Effect) and values_are_valid(ce.location_x, ce.location_y):
            return getters.get_units_in_area(self._uuid, ce.location_x, ce.location_y, ce.location_x, ce.location_y)

    def _get_trigger(self, ce: Union['Condition', 'Effect']) -> Optional[Trigger]:
        if value_is_valid(ce.trigger_id):
            return getters.get_trigger(self._uuid, ce.trigger_id)

    def _create_area(self, ce: Union['Condition', 'Effect']) -> Optional[Area]:
        if values_are_valid(ce.area_x1, ce.area_y1, ce.area_x2, ce.area_y2):
            return actions.new_area_object(self._uuid).select(ce.area_x1, ce.area_y1, ce.area_x2, ce.area_y2)

    def _create_tiles(self, ce: Union['Condition', 'Effect']) -> Optional[List[Tile]]:
        tiles = []
        if isinstance(ce, Effect) and values_are_valid(ce.location_x, ce.location_y):
            tiles.append(Tile(ce.location_x, ce.location_y))
        area = self._create_area(ce)
        if area:
            tiles.extend(list(area.to_coords()))
        return tiles

    def _get_unit_ids(self, ce: Union['Condition', 'Effect']) -> List[int]:
        ids: List[int] = []
        if isinstance(ce, Condition) and values_are_valid(ce.unit_object):
            ids.append(ce.unit_object)
        elif isinstance(ce, Effect):
            if value_is_valid(ce.location_object_reference):
                ids.append(ce.location_object_reference)
            if value_is_valid(ce.selected_object_ids):
                ids.extend(ce.selected_object_ids)
        return ids

    @staticmethod
    def _resolve_data_trigger_name(trigger: 'Trigger'):
        return trigger.name[trigger.name.find(":") + 1:]


def loop_trigger_content(trigger: 'Trigger') -> Iterator[Union['Condition', 'Effect']]:
    for condition in trigger.conditions:
        yield condition
    for effect in trigger.effects:
        yield effect



from __future__ import annotations

from typing import Dict, Type

from AoE2ScenarioParser.helper.printers import s_print
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.managers.de.map_manager_de import MapManagerDE
from AoE2ScenarioParser.objects.managers.de.trigger_manager_de import TriggerManagerDE
from AoE2ScenarioParser.objects.managers.de.unit_manager_de import UnitManagerDE
from AoE2ScenarioParser.objects.managers.de.xs_manager_de import XsManagerDE
from AoE2ScenarioParser.objects.managers.player_manager import PlayerManager
from AoE2ScenarioParser.scenarios.scenario_store import getters

managers: Dict[str, Dict[str, Type[AoE2Object]]] = {
    'DE': {
        'Player': PlayerManager,
        'Map': MapManagerDE,
        'Unit': UnitManagerDE,
        'Trigger': TriggerManagerDE,
        'Xs': XsManagerDE,
    }
}


class AoE2ObjectManager:
    def __init__(self, scenario_uuid):
        self.scenario_uuid = scenario_uuid
        self.managers = {}

    def setup(self):
        s_print(f"\nSetting up managers ...", final=True)
        gv = getters.get_game_version(self.scenario_uuid)

        for name, manager in managers[gv].items():
            s_print(f"\t🔄 Setting up {name}Manager...", color="yellow")
            self.managers[name] = manager._construct(self.scenario_uuid)
            s_print(f"\t✔ {name}Manager", final=True, color="green")

        s_print(f"Setting up managers finished successfully.", final=True)

    def reconstruct(self):
        s_print("\nReconstructing sections and structs from managers...", final=True)

        manager: AoE2Object
        for name, manager in self.managers.items():
            s_print(f"\t🔄 Reconstructing {name}Manager...", color="yellow")
            manager.commit()
            s_print(f"\t✔ {name}Manager", final=True, color="green")

        s_print("Reconstruction finished successfully.", final=True)



from enum import IntEnum


class ConditionId(IntEnum):
    """
    This enum class provides the integer values used to reference the conditions in the game. Used in every condition
    to indicate which type of condition it is

    **Examples**

    >>> ConditionId.BRING_OBJECT_TO_AREA
    >>> 1
    """
    NONE = 0
    """
    Attributes for the **none**  condition are:\n
    ... It's called none... What parameters do **you** think it has?! ❤ 
    """
    BRING_OBJECT_TO_AREA = 1
    """
    Attributes for the **bring_object_to_area** condition are:\n
    - unit_object  
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - inverted
    """
    BRING_OBJECT_TO_OBJECT = 2
    """
    Attributes for the **bring_object_to_object** condition are:\n
    - unit_object
    - next_object
    - inverted
    """
    OWN_OBJECTS = 3
    """
    Attributes for the **own_objects** condition are:\n
    - quantity
    - object_list
    - source_player
    - object_group
    - object_type
    - include_changeable_weapon_objects
    """
    OWN_FEWER_OBJECTS = 4
    """
    Attributes for the **own_fewer_objects** condition are:\n
    - quantity
    - object_list
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - include_changeable_weapon_objects
    """
    OBJECTS_IN_AREA = 5
    """
    Attributes for the **object_in_area** condition are:\n
    - quantity
    - object_list
    - source_player
    - area_x1
    - area_y1
    - area_x2
    - area_y2
    - object_group
    - object_type
    - object_state
    - inverted
    - include_changeable_weapon_objects
    """
    DESTROY_OBJECT = 6
    """
    Attributes for the **destroy_object** condition are:\n
    - unit_object
    - inverted
    """
    CAPTURE_OBJECT = 7
    """
    Attributes for the **capture_object** condition are:\n
    - unit_object
    - source_player
    - inverted
    """
    ACCUMULATE_ATTRIBUTE = 8
    """
    Attributes for the **accumulate_attribute** condition are:\n
    - quantity
    - attribute
    - source_player
    - inverted
    """
    RESEARCH_TECHNOLOGY = 9
    """
    Attributes for the **research_technology** condition are:\n
    - source_player
    - technology
    - inverted
    """
    TIMER = 10
    """
    Attributes for the **timer** condition are:\n
    - timer
    - inverted
    """
    OBJECT_SELECTED = 11
    """
    Attributes for the **object_selected** condition are:\n
    - unit_object
    - inverted
    """
    AI_SIGNAL = 12
    """
    Attributes for the **ai_signal** condition are:\n
    - ai_signal
    - inverted
    """
    PLAYER_DEFEATED = 13
    """
    Attributes for the **player_defeated** condition are:\n
    - source_player
    - inverted
    """
    OBJECT_HAS_TARGET = 14
    """
    Attributes for the **object_has_target** condition are:\n
    - unit_object
    - next_object
    - object_list
    - object_group
    - object_type
    - inverted
    """
    OBJECT_VISIBLE = 15
    """
    Attributes for the **object_visible** condition are:\n
    - unit_object
    """
    OBJECT_NOT_VISIBLE = 16
    """
    Attributes for the **object_not_visible** condition are:\n
    - unit_object
    """
    RESEARCHING_TECH = 17
    """
    Attributes for the **researching_tech** condition are:\n
    - source_player
    - technology
    - inverted
    """
    UNITS_GARRISONED = 18
    """
    Attributes for the **units_garrisoned** condition are:\n
    - quantity
    - unit_object
    - inverted
    """
    DIFFICULTY_LEVEL = 19
    """
    Attributes for the **difficulty_level** condition are:\n
    - condition_type
    - quantity
    - inverted
    """
    CHANCE = 20
    """
    Attributes for the **chance** condition are:\n
    - quantity
    """
    TECHNOLOGY_STATE = 21
    """
    Attributes for the **technology_state** condition are:\n
    - quantity
    - source_player
    - technology
    - inverted
    """
    VARIABLE_VALUE = 22
    """
    Attributes for the **variable_value** condition are:\n
    - quantity
    - inverted
    - variable
    - comparison
    """
    OBJECT_HP = 23
    """
    Attributes for the **object_hp** condition are:\n
    - quantity
    - unit_object
    - inverted
    - comparison
    """
    DIPLOMACY_STATE = 24
    """
    Attributes for the **diplomacy_state** condition are:\n
    - quantity
    - source_player
    - inverted
    - target_player
    """
    SCRIPT_CALL = 25
    """
    Attributes for the **script_call** condition are:\n
    - xs_function
    
    **Version notice**: \n
    This condition was added in: 1.40
    """
    OBJECT_SELECTED_MULTIPLAYER = 26
    """
    Attributes for the **object_selected_multiplayer** condition are:\n
    - unit_object
    - source_player
    - inverted
    
    **Version notice**: \n
    This condition was added in: 1.40
    """
    OBJECT_VISIBLE_MULTIPLAYER = 27
    """
    Attributes for the **object_visible_multiplayer** condition are:\n
    - unit_object
    - source_player
    - inverted
    
    **Version notice**: \n
    This condition was added in: 1.40
    """
    OBJECT_HAS_ACTION = 28
    """
    Attributes for the **object_has_action** condition are:\n
    - unit_object
    - next_object
    - inverted
    - unit_ai_action
    
    **Version notice**: \n
    This condition was added in: 1.40
    """
    OR = 29
    """
    The **OR** condition does not have any attributes
    
    **Version notice**: \n
    This condition was added in: 1.40
    """
    AI_SIGNAL_MULTIPLAYER = 30
    """
    Attributes for the **ai_signal_multiplayer** condition are:\n
    - ai_signal
    - inverted
    
    **Version notice**: \n
    This condition was added in: 1.40
    """
    BUILDING_IS_TRADING = 54
    """
    Attributes for the **building_is_trading** condition are:\n
    - unit_object
    - inverted
    
    **Version notice**: \n
    This condition was added in: 1.46
    """
    DISPLAY_TIMER_TRIGGERED = 55
    """
    Attributes for the **display_timer_triggered** condition are:\n
    - timer_id
    - inverted
    
    **Version notice**: \n
    This condition was added in: 1.46
    """
    VICTORY_TIMER = 56
    """
    Attributes for the **victory_timer** condition are:\n
    - quantity
    - source_player
    - inverted
    - comparison
    - victory_timer_type
    
    **Version notice**: \n
    This condition was added in: 1.46
    """
    AND = 57
    """
    The **AND** condition does not have any attributes
    
    **Version notice**: \n
    This condition was added in: 1.46
    """


empty_attributes = {
    "condition_type": -1,
    "quantity": -1,
    "attribute": -1,
    "unit_object": -1,
    "next_object": -1,
    "object_list": -1,
    "source_player": -1,
    "technology": -1,
    "timer": -1,
    "area_x1": -1,
    "area_y1": -1,
    "area_x2": -1,
    "area_y2": -1,
    "object_group": -1,
    "object_type": -1,
    "ai_signal": -1,
    "inverted": -1,
    "variable": -1,
    "comparison": -1,
    "target_player": -1,
    "unit_ai_action": -1,
    "object_state": -1,
    "timer_id": -1,
    "victory_timer_type": -1,
    "include_changeable_weapon_objects": -1,
    "xs_function": ""
}

# Set using the version json files
condition_names = {}
default_attributes = {}
attributes = {}
attribute_presentation = {}



from __future__ import annotations

from enum import IntEnum
from typing import List, Union


class PlayerId(IntEnum):
    @staticmethod
    def all(exclude_gaia: bool = False) -> List[PlayerId]:
        """
        Return a list of all players

        Args:
            exclude_gaia (bool): if the GAIA player should be excluded from the list or not

        Returns:
            The list of playerIds
        """
        return ([] if exclude_gaia else [PlayerId.GAIA]) + [
            PlayerId.ONE, PlayerId.TWO, PlayerId.THREE, PlayerId.FOUR,
            PlayerId.FIVE, PlayerId.SIX, PlayerId.SEVEN, PlayerId.EIGHT
        ]

    """
    This enum class provides the integer values used to reference each player in the game. Use these over specifying
    player number directly to improve code readability
    """
    GAIA = 0
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8


class PlayerColorId(IntEnum):
    """
    This enum class provides the integer values used to reference the colors in-game. These are used in the 'Change Object
    Color', etc. effects
    """
    BLUE = 1
    RED = 2
    GREEN = 3
    YELLOW = 4
    AQUA = 5
    PURPLE = 6
    GRAY = 7
    ORANGE = 8


class ColorId(IntEnum):
    """
    This enum represents the actual colors in the dropdown menu for players.
    Do not confuse with the PlayerColorId which is used to reference a player number by it's color.
    """
    BLUE = 0
    RED = 1
    GREEN = 2
    YELLOW = 3
    AQUA = 4
    PURPLE = 5
    GRAY = 6
    ORANGE = 7
    INVALID_8 = 8
    INVALID_9 = 9
    INVALID_10 = 10
    INVALID_11 = 11
    INVALID_12 = 12
    INVALID_13 = 13
    INVALID_14 = 14
    INVALID_15 = 15

    @staticmethod
    def from_player_id(player_id: Union[PlayerId, int]):
        if not 0 <= player_id <= 8:
            raise ValueError(f"Invalid player ID. Should be between 0 and 8 (both incl), but got: {player_id}")
        if player_id == 0:
            return ColorId.GRAY
        return ColorId(player_id - 1)



from AoE2ScenarioParser.datasets.dataset_enum import _DataSetIntEnums


class StartingAge(_DataSetIntEnums):
    DARK_AGE = 2
    FEUDAL_AGE = 3
    CASTLE_AGE = 4
    IMPERIAL_AGE = 5
    POST_IMPERIAL_AGE = 6


class Civilization(_DataSetIntEnums):
    BRITONS = 1
    FRANKS = 2
    GOTHS = 3
    TEUTONS = 4
    JAPANESE = 5
    CHINESE = 6
    BYZANTINES = 7
    PERSIANS = 8
    SARACENS = 9
    TURKS = 10
    VIKINGS = 11
    MONGOLS = 12
    CELTS = 13
    SPANISH = 14
    AZTECS = 15
    MAYANS = 16
    HUNS = 17
    KOREANS = 18
    ITALIANS = 19
    HINDUSTANIS = 20
    INCAS = 21
    MAGYARS = 22
    SLAVS = 23
    PORTUGUESE = 24
    ETHIOPIANS = 25
    MALIANS = 26
    BERBERS = 27
    KHMER = 28
    MALAY = 29
    BURMESE = 30
    VIETNAMESE = 31
    BULGARIANS = 32
    TATARS = 33
    CUMANS = 34
    LITHUANIANS = 35
    BURGUNDIANS = 36
    SICILIANS = 37
    POLES = 38
    BOHEMIANS = 39
    DRAVIDIANS = 40
    BENGALIS = 41
    GURJARAS = 42
    RANDOM = 43
    FULL_RANDOM = 45



from __future__ import annotations

from AoE2ScenarioParser.objects.managers.de.map_manager_de import MapManagerDE
from AoE2ScenarioParser.objects.managers.de.trigger_manager_de import TriggerManagerDE
from AoE2ScenarioParser.objects.managers.de.unit_manager_de import UnitManagerDE
from AoE2ScenarioParser.objects.managers.de.xs_manager_de import XsManagerDE
from AoE2ScenarioParser.objects.managers.player_manager import PlayerManager
from AoE2ScenarioParser.scenarios.aoe2_scenario import AoE2Scenario


class AoE2DEScenario(AoE2Scenario):
    @property
    def trigger_manager(self) -> TriggerManagerDE:
        return self._object_manager.managers['Trigger']

    @property
    def unit_manager(self) -> UnitManagerDE:
        return self._object_manager.managers['Unit']

    @property
    def map_manager(self) -> MapManagerDE:
        return self._object_manager.managers['Map']

    @property
    def xs_manager(self) -> XsManagerDE:
        return self._object_manager.managers['Xs']

    @property
    def player_manager(self) -> PlayerManager:
        return self._object_manager.managers['Player']

    @classmethod
    def from_file(cls, filename, game_version="DE") -> AoE2DEScenario:
        return super().from_file(filename, game_version)



from __future__ import annotations

from typing import Dict

from AoE2ScenarioParser.helper.pretty_format import pretty_format_dict
from AoE2ScenarioParser.sections.retrievers.retriever import Retriever


class AoE2StructModel:
    def __init__(self, name: str, retriever_map: Dict[str, Retriever], structs: Dict[AoE2StructModel]):
        self.name = name
        self.retriever_map = retriever_map
        self.structs = structs

    @classmethod
    def from_structure(cls, name, structure) -> AoE2StructModel:
        retriever_map = {}
        for retriever_name, attr in structure.get('retrievers').items():
            retriever_map[retriever_name] = Retriever.from_structure(retriever_name, attr)
        structs = model_dict_from_structure(structure)

        return cls(name, retriever_map, structs)

    def __str__(self):
        return f"[AoE2StructModel] {self.name} -> retrievers: " + pretty_format_dict(self.retriever_map)


def model_dict_from_structure(structure) -> Dict[AoE2StructModel]:
    models = {}
    for name, attr in structure.get('structs', {}).items():
        # Create struct model
        models[name] = AoE2StructModel.from_structure(name, attr)
    return models



from __future__ import annotations

import pickle
from typing import Dict

from AoE2ScenarioParser.helper import bytes_parser, string_manipulations
from AoE2ScenarioParser.helper.bytes_conversions import parse_bytes_to_val, parse_val_to_bytes
from AoE2ScenarioParser.helper.list_functions import listify
from AoE2ScenarioParser.helper.pretty_format import pretty_format_list
from AoE2ScenarioParser.sections.dependencies.dependency_action import DependencyAction
from AoE2ScenarioParser.sections.dependencies.retriever_dependency import RetrieverDependency
from AoE2ScenarioParser.sections.retrievers.datatype import DataType


class Retriever:
    """ A Class for defining how to retrieve data.
    The Constructor has quite some parameters which can all be used for getting the proper data
    """

    __slots__ = [
        'on_construct',
        'on_commit',
        'on_refresh',
        'name',
        'datatype',
        'is_list',
        'log_value',
        '_data',
        'default_value',
    ]

    on_construct: RetrieverDependency
    on_commit: RetrieverDependency
    on_refresh: RetrieverDependency

    def __init__(self, name, default_value=None, datatype=DataType(), is_list=None, log_value=False):
        """
        Args:
            name (str): The name of the item. Has to be unique within the Section or Struct
            default_value (Any): The default value of this retriever
            datatype (DataType): A datatype object
            is_list (Union[None, bool]): If this retriever data should be presented using a list. If None, it's unknown.
            log_value (bool): A boolean for, mostly, debugging. This will log information about this retrievers when the
                data is changed, when this retriever is constructed and when it's committed.
        """
        self.name: str = name
        self.default_value = default_value
        self.datatype: DataType = datatype
        self.is_list = is_list
        self.log_value = log_value
        self._data = None

        if log_value:
            self.datatype.log_value = True
            self.datatype._debug_retriever_name = name

    def get_data_as_bytes(self):
        self.update_datatype_repeat()

        if self.data is not None and self.datatype.repeat != 0:
            result = []
            if self.datatype.type == "struct":
                for struct in self.data:
                    result.append(struct.get_data_as_bytes())
            else:
                for value in listify(self.data):
                    result.append(parse_val_to_bytes(self, value))

            joined_result = b''.join(result)
        else:
            joined_result = b''

        if self.log_value:
            print(f"{self.to_simple_string()} (Data: {self.data}) retrieved: {joined_result}")

        return joined_result

    def set_data_from_bytes(self, bytes_list):
        if self.datatype.repeat > 0 and len(bytes_list) == 0:
            raise ValueError("Unable to set bytes when no bytes are given")
        if self.datatype.repeat > 0 and self.datatype.repeat != len(bytes_list):
            raise ValueError("Unable to set bytes when bytes list isn't equal to repeat")

        result = [parse_bytes_to_val(self, entry_bytes) for entry_bytes in bytes_list]
        self.data = bytes_parser.vorl(self, result)

    def update_datatype_repeat(self):
        if type(self.data) == list:
            self.datatype.repeat = len(self.data)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if self.log_value:
            old_value = self._data
            self._print_value_update(old_value, value)

        # If repeat is 0 and value being said is truthy (mainly not an empty list) set repeat to one
        if self.datatype.repeat == 0 and value:
            self.datatype.repeat = 1
        self._data = value

    def set_data_to_default(self):
        if self.datatype.type == "data":
            data = bytes.fromhex(self.default_value)
        elif type(self.default_value) is list:
            data = self.default_value.copy()
            assert data is not self.default_value
        else:
            data = self.default_value

        self.data = data

    def duplicate(self):
        retriever = Retriever(
            name=self.name,
            default_value=self.default_value,
            datatype=self.datatype.duplicate(),
            is_list=self.is_list,
            log_value=self.log_value
        )
        for attr in attributes:
            if hasattr(self, attr):
                setattr(retriever, attr, getattr(self, attr))
        return retriever

    @classmethod
    def from_structure(cls, name, structure):
        datatype = DataType(var=structure.get('type'), repeat=structure.get('repeat', 1))
        retriever = cls(
            name=name,
            default_value=structure.get('default'),
            datatype=datatype,
            is_list=structure.get('is_list', None),
            log_value=structure.get('log', False)
        )
        # Go through dependencies if exist, else empty dict
        for dependency_name, properties in structure.get('dependencies', {}).items():
            if type(properties) is not list:
                r_dep = RetrieverDependency.from_structure(properties)
                _evaluate_is_list_attribute(retriever, r_dep)
                setattr(retriever, dependency_name, r_dep)
            else:
                dependency_list = []
                for dependency_struct in properties:
                    r_dep = RetrieverDependency.from_structure(dependency_struct)
                    _evaluate_is_list_attribute(retriever, r_dep)
                    dependency_list.append(r_dep)
                setattr(retriever, dependency_name, dependency_list)
        return retriever

    def _print_value_update(self, old, new) -> None:
        """
        Function to print when data is changed. Can also be called for when data is changed but the property doesn't
        fire. This happens when an array is adjusted in size by appending to it ([...] += [...]).

        Args:
            old (str): The old value represented using a string
            new (str): The new value represented using a string
        """
        print(f"{self.to_simple_string()} >>> set to: {string_manipulations.q_str(new)} "
              f"(was: {string_manipulations.q_str(old)})")

    def get_short_str(self):
        if self.data is not None:
            if type(self.data) is list:
                data = pretty_format_list(self.data)
            else:
                data = string_manipulations.q_str(self.data)
        else:
            data = "<None>"
        return f"{self.name} ({self.datatype.to_simple_string()}): {data}"

    def to_simple_string(self):
        return f"[Retriever] {self.name}: {self.datatype} (Default: {string_manipulations.q_str(self.default_value)})"

    def __repr__(self):
        if type(self.data) is list:
            data = str(pretty_format_list(self.data))
        else:
            data = string_manipulations.q_str(self.data)
        return f"{self.to_simple_string()} >>> {data}"


def duplicate_retriever_map(retriever_map: Dict[str, Retriever]) -> Dict[str, Retriever]:
    return pickle.loads(pickle.dumps(retriever_map))


def reset_retriever_map(retriever_map: Dict[str, Retriever]) -> None:
    for retriever in retriever_map.values():
        retriever.set_data_to_default()


def _evaluate_is_list_attribute(retriever, dependency):
    if dependency.dependency_action == DependencyAction.SET_REPEAT and retriever.is_list is None:
        retriever.is_list = True


attributes = ['on_refresh', 'on_construct', 'on_commit']



from enum import Enum, IntEnum, IntFlag, EnumMeta
from typing import Union, Type


class _DataSetMeta(EnumMeta):
    def __contains__(self, other):
        try:
            self(other)
        except ValueError:
            return False
        else:
            return True


class _DataSet(Enum, metaclass=_DataSetMeta):
    def attribute_presentation(self) -> str:
        raise NotImplemented("_DataSet.attribute_presentation has to be implemented")


class _DataSetIntEnums(_DataSet, IntEnum):
    def attribute_presentation(self) -> str:
        """
        Get the string representation of an enum entry. Uses `.name` when not overridden.
        Returns:
            The string representation of an enum entry.
        """
        return super().name


class _DataSetIntFlags(_DataSet, IntFlag):
    def attribute_presentation(self) -> str:
        """
        Get the string representation of an enum entry. Uses `.name` when not overridden.
        Returns:
            The string representation of an enum entry.
        """
        return super().name


def dataset_or_value(enum: Type[_DataSet], value: Union[int, str]) -> Union[_DataSet, int, str]:
    """
    Return the value in the enum used to create the enum, or if it failed, returns just the value

    Args:
        enum: The enum to create with the given value
        value: The value to use in the enum

    Returns:
        The enum with the given value if it exists, the value itself otherwise
    """
    try:
        return enum(value)
    except ValueError:
        return value



from typing import Type, List
from uuid import UUID

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.exceptions import UnsupportedAttributeError
from AoE2ScenarioParser.helper.helper import exclusive_if
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.support.uuid_list import NO_UUID
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection
from AoE2ScenarioParser.sections.dependencies.dependency import handle_retriever_dependency
from AoE2ScenarioParser.sections.retrievers.support import Support


class RetrieverObjectLink:
    def __init__(
            self,
            variable_name: str,
            section_name: str = None,
            link: str = None,
            support: Support = None,
            process_as_object: Type[AoE2Object] = None,
            retrieve_history_number: int = -1,
            commit_callback=None,
            destination_object=None,
    ):
        if not exclusive_if(link is not None, retrieve_history_number != -1):
            raise ValueError("You must use exactly one of 'link' and the two 'retrieve...number' parameters.")

        self.name: str = variable_name
        self.section_name = section_name
        self.link = link
        self.support: Support = support
        self.process_as_object: Type[AoE2Object] = process_as_object
        self.retrieve_history_number: int = retrieve_history_number
        self.commit_callback = commit_callback
        self.destination_object = destination_object

        self.splitted_link: List[str] = link.split('.') if link is not None else []

    def construct(self, host_uuid, number_hist=None):
        if number_hist is None:
            number_hist = []

        if self.retrieve_history_number != -1:
            return number_hist[self.retrieve_history_number]
        else:
            sections = getters.get_sections(host_uuid)

            # Retrieve value without using eval() -- Eval is slow
            value = sections[self.section_name]
            for index, item in enumerate(self.splitted_link):
                if item.endswith("]"):
                    # item[:-11] removes "[__index__]" from the key
                    value = getattr(value, item[:-11])[number_hist[index]]
                else:
                    value = getattr(value, item)

            if self.process_as_object:
                return self.process_object_list(value, number_hist, host_uuid)
            return value

    def process_object_list(self, value_list, instance_number_history, host_uuid):
        object_list = []
        for index, struct in enumerate(value_list):
            object_list.append(
                self.process_as_object._construct(host_uuid, instance_number_history + [index])
            )
        return object_list

    def commit(self, host_uuid: UUID, host_obj: AoE2Object):
        # Object-only attributes for the ease of access of information.
        # Not actually representing a value in the scenario file.
        if self.retrieve_history_number != -1:
            return

        if host_uuid == NO_UUID:
            raise ValueError(f"Invalid object commit. No UUID was set. Object class: {host_obj.__class__.__name__}")

        number_hist = host_obj._instance_number_history

        try:
            # Get new value for receiver
            value = getattr(host_obj, self.name)
        except UnsupportedAttributeError:
            return  # Not supported in current version.

        if self.commit_callback is not None:
            value = self.commit_callback(host_obj, self.name, value)

        sections = getters.get_sections(host_uuid)
        section = sections[self.section_name]

        # Retrieve value without using eval() -- Eval is slow
        retriever = None
        file_section = section
        for index, item in enumerate(self.splitted_link):
            try:
                if "[" in item:
                    file_section = getattr(file_section, item[:-11])[number_hist[index]]
                else:
                    retriever = file_section.retriever_map[item]
            except KeyError as e:
                print(e)
                # Maybe not supported in current version. if actually not supported -> ignore
                if self.support is not None:
                    if not self.support.supports(
                            getters.get_scenario_version(host_uuid)):
                        return
                if settings.IGNORE_WRITING_ERRORS:
                    return
                raise e

        if retriever is None:
            raise ValueError("RetrieverObjectLink is unable to find retriever")

        if self.process_as_object:
            struct_datatype = retriever.datatype.var

            prefix = "struct:"
            if not struct_datatype.startswith(prefix):
                raise ValueError(
                    f"process_as_object isn't defined properly. Expected: '{prefix}...', got: '{struct_datatype}'"
                )

            struct_name = struct_datatype[len(prefix):]
            struct_model = file_section.struct_models[struct_name]

            RetrieverObjectLink.update_retriever_length(retriever, struct_model, len(value), host_uuid)
            RetrieverObjectLink.commit_object_list(value, host_obj._instance_number_history)
        else:
            retriever.data = value

        if hasattr(retriever, 'on_commit'):
            handle_retriever_dependency(retriever, "commit", file_section, host_uuid)

    @staticmethod
    def commit_object_list(object_list, instance_number_history):
        obj: AoE2Object
        for index, obj in enumerate(object_list):
            obj._instance_number_history = instance_number_history + [index]
            obj.commit()

    @staticmethod
    def update_retriever_length(retriever, model, new_len, host_uuid):
        try:
            old_len = len(retriever.data)
        except TypeError:  # retriever.data was not set before (list of 0 -> None)
            old_len = 0
            retriever.data = []

        if new_len < old_len:
            retriever.data = retriever.data[:new_len]
        elif new_len > old_len:
            retriever.data += [
                AoE2FileSection.from_model(model, host_uuid, set_defaults=True)
                for _ in range(new_len - old_len)
            ]

            if retriever.log_value:
                retriever._print_value_update(f"[{model.name}] * {old_len}", f"[{model.name}] * {new_len}")

    def __repr__(self):
        return f"[RetrieverObjectLink] {self.name}: {self.section_name}.{self.link}" + \
               (f"\n\t- Process as: {self.process_as_object.__name__}" if self.process_as_object else "") + \
               (f"\n\t- Get Hist Number: {self.retrieve_history_number}" if self.retrieve_history_number >= 0 else "")



from __future__ import annotations

from copy import deepcopy
from enum import Enum
from typing import List, Type, TYPE_CHECKING

from AoE2ScenarioParser.helper.exceptions import UnsupportedAttributeError
from AoE2ScenarioParser.helper.pretty_format import pretty_format_dict
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.support.uuid_list import NO_UUID
from AoE2ScenarioParser.scenarios.scenario_store import getters

if TYPE_CHECKING:
    from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class AoE2Object:
    _link_list: List[RetrieverObjectLink] = []

    def __init__(self, **kwargs):
        self._instance_number_history = []
        self._host_uuid = kwargs.get('host_uuid', NO_UUID)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            entry = self._deepcopy_entry(k, v)
            if entry is None:
                continue
            setattr(result, k, entry)
        return result

    def _deepcopy_entry(self, k, v):
        if k in ['_sections', '_link_list']:
            val = getattr(self, k)
        else:
            val = deepcopy(v)
        return val

    @classmethod
    def _construct(cls, host_uuid, number_hist=None):
        if number_hist is None:
            number_hist = []

        scenario_version = getters.get_scenario_version(host_uuid)

        object_parameters = {}
        for link in cls._link_list:
            if link.support is not None and not link.support.supports(scenario_version):

                error_msg = f_unsupported_string(link, scenario_version)

                def _get(self):
                    raise UnsupportedAttributeError(error_msg)

                def _set(self, val):
                    if val is not None:
                        raise UnsupportedAttributeError(error_msg)

                obj = cls
                if link.destination_object is not None:
                    obj = link.destination_object

                # Todo: Runs for each _construct() -- A LOT of overhead
                #  Doesn't work properly when reading an older scenario first, and a newer one later
                #  Properties don't get reset!
                setattr(obj, link.name, property(_get, _set))
                object_parameters[link.name] = None
            else:
                object_parameters[link.name] = link.construct(host_uuid, number_hist=number_hist)

        object_parameters['host_uuid'] = host_uuid

        return cls(**object_parameters)

    def commit(self, local_link_list=None):
        """
        Commits all changes to the section & struct structure of the object it's called upon.

        Args:
            local_link_list (Type[List[RetrieverObjectLink]]): a separate list of RetrieverObjectLinks. This way it's
                possible to commit only specific properties instead of all from an object.
        """
        if local_link_list is None:
            local_link_list = self._link_list

        for link in local_link_list[::-1]:
            link.commit(self._host_uuid, host_obj=self)

    @staticmethod
    def get_instance_number(obj: AoE2Object = None, number_hist=None) -> int:
        if obj is None and number_hist is None:
            raise ValueError("The use of the parameter 'obj' or 'number_hist' is required.")
        if obj is not None and number_hist is not None:
            raise ValueError("Cannot use both the parameter 'obj' and 'number_hist'.")

        if number_hist is None and obj is not None:
            number_hist = obj._instance_number_history
        return number_hist[-1] if len(number_hist) > 0 else None

    def _get_object_attrs(self):
        attrs = ["_instance_number_history", "_host_uuid"]
        for link in self._link_list:
            attrs.append(link.name)
        return attrs

    def __repr__(self):
        attrs = self._get_object_attrs()

        self_dict = {}
        for attr in attrs:
            value = getattr(self, attr)
            if isinstance(value, Enum):
                self_dict[attr] = value.name
            else:
                self_dict[attr] = value

        return str(self.__class__.__name__) + ": " + add_tabs(pretty_format_dict(self_dict), 1)


def f_unsupported_string(link: RetrieverObjectLink, version: str):
    return f"The property '{link.name}' is {link.support}. Current version: {version}.\n"



from __future__ import annotations

import copy
from enum import IntEnum
from typing import List, Dict, Union

from AoE2ScenarioParser.datasets.effects import EffectId
from AoE2ScenarioParser.datasets.players import PlayerId
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.helper import value_is_valid
from AoE2ScenarioParser.helper.list_functions import hash_list, list_changed, update_order_array
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.effect import Effect
from AoE2ScenarioParser.objects.data_objects.trigger import Trigger
from AoE2ScenarioParser.objects.support.enums.group_by import GroupBy
from AoE2ScenarioParser.objects.support.trigger_select import TriggerSelect, TS
from AoE2ScenarioParser.objects.support.uuid_list import UuidList
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class TriggerManager(AoE2Object):
    """Manager of the everything trigger related."""

    _link_list = [
        RetrieverObjectLink("triggers", "Triggers", "trigger_data", process_as_object=Trigger),
        RetrieverObjectLink("trigger_display_order", "Triggers", "trigger_display_order_array"),
    ]

    def __init__(self,
                 triggers: List[Trigger],
                 trigger_display_order: List[int],
                 **kwargs
                 ):
        super().__init__(**kwargs)

        self._trigger_hash = hash_list(triggers)
        self.triggers: List[Trigger] = triggers
        self.trigger_display_order: List[int] = trigger_display_order

    @property
    def triggers(self) -> List[Trigger]:
        return self._triggers

    @triggers.setter
    def triggers(self, value):
        value = UuidList(self._host_uuid, value, on_update_execute_entry=self._update_triggers_uuid)

        self._trigger_hash = hash_list(value)
        self._triggers = value
        self.trigger_display_order = list(range(len(value)))

    def _update_triggers_uuid(self, trigger):
        """Function to update inner UUIDs """
        for effect in trigger.effects:
            effect._host_uuid = self._host_uuid
        for condition in trigger.conditions:
            condition._host_uuid = self._host_uuid

    @property
    def trigger_display_order(self) -> List[int]:
        if list_changed(self.triggers, self._trigger_hash):
            update_order_array(self._trigger_display_order, len(self.triggers))
            self._trigger_hash = hash_list(self.triggers)
        return self._trigger_display_order

    @trigger_display_order.setter
    def trigger_display_order(self, val):
        self._trigger_display_order = val

    def copy_trigger_per_player(self,
                                from_player,
                                trigger_select: Union[int, TriggerSelect],
                                change_from_player_only=False,
                                include_player_source=True,
                                include_player_target=False,
                                trigger_ce_lock=None,
                                include_gaia: bool = False,
                                create_copy_for_players: List[IntEnum] = None) -> Dict[PlayerId, Trigger]:
        """
        Copies a trigger for all or a selection of players. Every copy will change desired player attributes with it.

        Args:
            from_player (IntEnum): The central player this trigger is created for. This is the player that will not get
                a copy.
            trigger_select (Union[int, TriggerSelect]): An object used to identify which trigger to select.
            change_from_player_only (bool): If set to True, only change player attributes in effects and conditions that
                are equal to the player defined using the `from_player` parameter.
            include_player_source (bool): If set to True, allow player source attributes to be changed while copying.
                Player source attributes are attributes where a player is defined to perform an action such as create an
                object. If set to False these attributes will remain unchanged.
            include_player_target (bool): If set to True, allow player target attributes to be changed while copying.
                Player target attributes are attributes where a player is defined as the target such as change ownership
                or sending resources. If set to False these attributes will remain unchanged.
            trigger_ce_lock (TriggerCELock): The TriggerCELock object. Used to lock certain (types) of conditions or
                effects from being changed while copying.
            include_gaia (bool): If True, GAIA is included in the copied list. (Also when `create_copy_for_players` is
                defined)
            create_copy_for_players (List[IntEnum]): A list of Players to create a copy for. The `from_player` will be
                excluded from this list.

        Returns:
            A dict with all the new created triggers. The key is the player for which the trigger is
                created using the IntEnum associated with it. Example:
                {PlayerId.TWO: Trigger, PlayerId.FIVE: Trigger}

        Raises:
            ValueError: if more than one trigger selection is used. Any of (trigger_index, display_index or trigger)
                Or if Both `include_player_source` and `include_player_target` are `False`

        :Authors:
            KSneijders

        """
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        if create_copy_for_players is None:
            create_copy_for_players = [
                PlayerId.ONE, PlayerId.TWO, PlayerId.THREE, PlayerId.FOUR,
                PlayerId.FIVE, PlayerId.SIX, PlayerId.SEVEN, PlayerId.EIGHT
            ]
        if include_gaia and PlayerId.GAIA not in create_copy_for_players:
            create_copy_for_players.append(PlayerId.GAIA)

        alter_conditions, alter_effects = TriggerManager._find_alterable_ce(trigger, trigger_ce_lock)

        return_dict: Dict[PlayerId, Trigger] = {}
        for player in create_copy_for_players:
            if player == from_player:
                continue

            new_trigger = self.copy_trigger(TS.trigger(trigger), append_after_source=False, add_suffix=False)
            new_trigger.name += " (GAIA)" if player == PlayerId.GAIA else f" (p{player})"
            return_dict[player] = new_trigger

            for cond_x in alter_conditions:
                cond = new_trigger.conditions[cond_x]
                if cond.source_player == -1:
                    continue

                if include_player_source:
                    if not change_from_player_only or (change_from_player_only and cond.source_player == from_player):
                        cond.source_player = PlayerId(player)
                if include_player_target:
                    if not change_from_player_only or (change_from_player_only and cond.target_player == from_player):
                        cond.target_player = PlayerId(player)

            for effect_x in alter_effects:
                effect = new_trigger.effects[effect_x]
                if effect.source_player == -1:
                    continue

                if include_player_source:
                    if not change_from_player_only or (change_from_player_only and effect.source_player == from_player):
                        effect.source_player = PlayerId(player)
                if include_player_target:
                    if not change_from_player_only or (change_from_player_only and effect.target_player == from_player):
                        effect.target_player = PlayerId(player)

        # After copies have been made
        trigger.name += f" (p{from_player})"

        return return_dict

    def copy_trigger(
            self,
            trigger_select: Union[int, TriggerSelect],
            append_after_source=True,
            add_suffix=True
    ) -> Trigger:
        """
        Creates an exact copy (deepcopy) of this trigger.

        Args:
            trigger_select (Union[int, TriggerSelect]): An object used to identify which trigger to select.
            append_after_source (bool): If the new trigger should be appended below the source trigger
            add_suffix (bool): If the text ' (copy)' should be added after the trigger

        Returns:
            The newly copied trigger
        """
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        deepcopy_trigger = copy.deepcopy(trigger)
        deepcopy_trigger.trigger_id = len(self.triggers)
        if add_suffix:
            deepcopy_trigger.name += " (copy)"

        self.triggers.append(deepcopy_trigger)

        if append_after_source:
            self.move_triggers([trigger_index, deepcopy_trigger.trigger_id], trigger_index)

        return deepcopy_trigger

    def copy_trigger_tree_per_player(self,
                                     from_player,
                                     trigger_select: Union[int, TriggerSelect],
                                     change_from_player_only=False,
                                     include_player_source=True,
                                     include_player_target=False,
                                     trigger_ce_lock=None,
                                     include_gaia: bool = False,
                                     create_copy_for_players: List[IntEnum] = None,
                                     group_triggers_by=None):
        """
        Copies an entire trigger tree for all or a selection of players. Every copy will change desired player
        attributes with it. Trigger trees are triggers linked together using EffectId.(DE)ACTIVATE_TRIGGER.

        Args:
            from_player (IntEnum): The central player this trigger is created for. This is the player that will not get
                a copy.
            trigger_select (Union[int, TriggerSelect]): An object used to identify which trigger to select.
            change_from_player_only (bool): If set to True, only change player attributes in effects and conditions that
                are equal to the player defined using the `from_player` parameter.
            include_player_source (bool): If set to True, allow player source attributes to be changed while copying.
                Player source attributes are attributes where a player is defined to perform an action such as create an
                object. If set to False these attributes will remain unchanged.
            include_player_target (bool): If set to True, allow player target attributes to be changed while copying.
                Player target attributes are attributes where a player is defined as the target such as change ownership
                or sending resources. If set to False these attributes will remain unchanged.
            trigger_ce_lock (TriggerCELock): The TriggerCELock object. Used to lock certain (types) of conditions or
                effects from being changed while copying.
            include_gaia (bool): If True, GAIA is included in the copied list. (Also when `create_copy_for_players` is
                defined)
            create_copy_for_players (List[IntEnum]): A list of Players to create a copy for. The `from_player` will be
                excluded from this list.
            group_triggers_by (GroupBy): How to group the newly added triggers.

        Returns:
            The newly created triggers in a dict using the Player as key and as value with a list of triggers
        """
        if group_triggers_by is None:
            group_triggers_by = GroupBy.NONE

        trigger_index, display_index, source_trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        known_node_indexes = [trigger_index]
        self._find_trigger_tree_nodes_recursively(source_trigger, known_node_indexes)

        new_triggers = {}
        trigger_index_swap = {}

        # Set values for from_player
        new_triggers[from_player] = [self.triggers[i] for i in known_node_indexes]
        for index in known_node_indexes:
            trigger = self.triggers[index]
            trigger_index_swap.setdefault(index, {})[from_player] = trigger.trigger_id

        # Copy for all other players
        for index in known_node_indexes:
            triggers = self.copy_trigger_per_player(
                from_player,
                index,
                change_from_player_only,
                include_player_source,
                include_player_target,
                trigger_ce_lock,
                include_gaia,
                create_copy_for_players,
            )
            for player, trigger in triggers.items():
                trigger_index_swap.setdefault(index, {})[player] = trigger.trigger_id
                new_triggers.setdefault(player, []).append(trigger)

        # Set trigger_id's in activation effects to the new player copied trigger ID
        for player, triggers in new_triggers.items():
            for trigger in triggers:
                for effect in get_activation_effects(trigger):
                    effect.trigger_id = trigger_index_swap[effect.trigger_id][player]

        # -------------- Group by logic -------------- #
        new_trigger_ids = []
        if group_triggers_by == GroupBy.TRIGGER:
            for i in range(len(known_node_indexes)):
                for player in PlayerId.all():
                    if player == from_player:
                        new_trigger_ids.append(known_node_indexes[i])
                        continue
                    if player not in new_triggers:
                        continue
                    new_trigger_ids.append(new_triggers[player][i].trigger_id)
        elif group_triggers_by == GroupBy.PLAYER:
            for player in PlayerId.all():
                if player == from_player:
                    new_trigger_ids.extend(known_node_indexes)
                    continue
                if player not in new_triggers:
                    continue

                new_trigger_ids.extend([trigger.trigger_id for trigger in new_triggers[player]])

        if group_triggers_by != GroupBy.NONE:
            self.move_triggers(new_trigger_ids, display_index)

        return new_triggers

    def move_triggers(self, trigger_ids: List[int], insert_index: int) -> None:
        """
        Function to move triggers. Moves the given IDs from anywhere to the split index. This function reorders triggers
        BUT keeps ``(de)activate trigger`` effects linked properly!

        As an example:

        >>> [0,1,2,3,4,5,6,7,8]  # Current index order
        >>> # Let's move trigger 1, 4, 5 and 6 to location 2
        >>> self.move_triggers([1, 4, 5, 6], 2)  # << 2 is an INDEX, not the value
        >>> [0,1,4,5,6,2,3,7,8]  # New index order

        Args:
            trigger_ids: The trigger IDs to move
            insert_index: The index that defines where to insert the triggers
        """
        if min(trigger_ids) < 0:
            raise ValueError(f"Trigger IDs cannot be negative")

        if insert_index >= len(self.trigger_display_order):
            # Add to the end of the list
            new_trigger_id_order = [n for n in self.trigger_display_order if n not in trigger_ids]
            new_trigger_id_order += trigger_ids
        else:
            insert_num = self.trigger_display_order[insert_index]
            new_trigger_id_order = [n for n in self.trigger_display_order if n not in trigger_ids or n == insert_num]

            split_index = new_trigger_id_order.index(insert_num)

            if insert_num in trigger_ids:
                new_trigger_id_order.remove(insert_num)

            new_trigger_id_order = new_trigger_id_order[:split_index] + trigger_ids + new_trigger_id_order[split_index:]
        self.reorder_triggers(new_trigger_id_order)

    def reorder_triggers(self, new_id_order: List[int] = None):
        """
        Reorder all triggers to a given order of IDs. This function reorders triggers BUT keeps ``(de)activate trigger``
        effects linked properly!

        As an example:

        >>> [0,1,2,3,4,5,6,7,8]  # Current index order
        >>> self.reorder_triggers([0,1,2,3,5,4,7,8,6])
        >>> [0,1,2,3,5,4,7,8,6]  # New index order

        Keep in mind that all trigger IDs will get remapped with this function. So ``trigger_manager.triggers[4]`` might
        result in a different trigger after this function is called in comparison to before.

        Args:
            new_id_order: The new trigger order. Uses the current display order when left unused
        """
        if new_id_order is not None:
            if min(new_id_order) < 0:
                raise ValueError(f"Trigger IDs cannot be negative")
            self.trigger_display_order = new_id_order

        new_triggers_list = []
        index_changes = {}
        for new_index, index in enumerate(self.trigger_display_order):
            try:
                trigger = self.triggers[index]
            except IndexError:
                raise ValueError(f"The trigger ID {index} doesn't exist") from None
            index_changes[trigger.trigger_id] = new_index

            trigger.trigger_id = new_index
            new_triggers_list.append(trigger)
        self.triggers = new_triggers_list

        # Find and update all (de)activation effect trigger references
        for trigger in self.triggers:
            for effect in get_activation_effects(trigger):
                if effect.trigger_id in index_changes:
                    effect.trigger_id = index_changes[effect.trigger_id]

    def copy_trigger_tree(self, trigger_select: Union[int, TriggerSelect]) -> List[Trigger]:
        """
        Copies an entire trigger tree. Trigger trees are triggers linked together using EffectId.(DE)ACTIVATE_TRIGGER.

        Args:
            trigger_select: An object used to identify which trigger to select

        Returns:
            The newly created triggers in a list
        """
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        known_node_indexes = [trigger_index]
        self._find_trigger_tree_nodes_recursively(trigger, known_node_indexes)

        new_triggers = []
        id_swap = {}
        for index in known_node_indexes:
            trigger = self.copy_trigger(index, append_after_source=False)
            new_triggers.append(trigger)
            id_swap[index] = trigger.trigger_id

        for trigger in new_triggers:
            activation_effects = [
                effect for effect in trigger.effects if
                effect.effect_type in [EffectId.ACTIVATE_TRIGGER, EffectId.DEACTIVATE_TRIGGER]
            ]
            for effect in activation_effects:
                effect.trigger_id = id_swap[effect.trigger_id]

        return new_triggers

    def replace_player(self, trigger_select: Union[int, TriggerSelect], to_player, only_change_from=None,
                       include_player_source=True, include_player_target=False, trigger_ce_lock=None) -> Trigger:
        """
        Replaces player attributes. Specifically useful if multiple players are used in the same trigger.

        Args:
            trigger_select (Union[int, TriggerSelect]): An object used to identify which trigger to select.
            to_player (PlayerId): The player the attributes are changed to.
            only_change_from (PlayerId): Can only change player attributes if the player is equal to the given value
            include_player_source (bool): If set to True, allow player source attributes to be changed while replacing.
                Player source attributes are attributes where a player is defined to perform an action such as create an
                object. If set to False these attributes will remain unchanged.
            include_player_target (bool): If set to True, allow player target attributes to be changed while replacing.
                Player target attributes are attributes where a player is defined as the target such as change ownership
                or sending resources. If set to False these attributes will remain unchanged.
            trigger_ce_lock (TriggerCELock): The TriggerCELock object. Used to lock certain (types) of conditions or
                effects from being changed.

        Returns:
            The given trigger with the proper player attributes changed
        """
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)
        alter_conditions, alter_effects = TriggerManager._find_alterable_ce(trigger, trigger_ce_lock)

        for cond_x in alter_conditions:
            cond = trigger.conditions[cond_x]
            if value_is_valid(cond.source_player) and include_player_source:
                if only_change_from is not None and only_change_from != cond.source_player:
                    continue
                cond.source_player = PlayerId(to_player)
            if value_is_valid(cond.target_player) and include_player_target:
                if only_change_from is not None and only_change_from != cond.target_player:
                    continue
                cond.target_player = PlayerId(to_player)
        for effect_x in alter_effects:
            effect = trigger.effects[effect_x]
            if value_is_valid(effect.source_player) and include_player_source:
                if only_change_from is not None and only_change_from != effect.source_player:
                    continue
                effect.source_player = PlayerId(to_player)
            if value_is_valid(effect.target_player) and include_player_target:
                if only_change_from is not None and only_change_from != effect.target_player:
                    continue
                effect.target_player = PlayerId(to_player)

        return trigger

    def add_trigger(self, name, description=None, description_stid=None, display_as_objective=None,
                    short_description=None, short_description_stid=None, display_on_screen=None, description_order=None,
                    enabled=None, looping=None, header=None, mute_objectives=None, conditions=None,
                    effects=None) -> Trigger:
        """
        Adds a new trigger to the scenario.

        Args:
            name (str): The name for the trigger
            description (str): The trigger description
            description_stid (int): The trigger description string table ID
            display_as_objective (bool): Display the trigger as objective
            short_description (str): The short trigger description
            short_description_stid (int): The short trigger description string table ID
            display_on_screen (bool): Display the trigger objective on screen
            description_order (int): ?
            enabled (bool): If the trigger is enabled from the start.
            looping (bool): If the trigger loops.
            header (bool): Turn objective into header
            mute_objectives (bool): Mute objectives
            conditions (List): A list of condition managers
            effects (List): A list of effect managers

        Returns:
            The newly created trigger
        """
        keys = [
            'description', 'description_stid', 'display_as_objective', 'short_description',
            'short_description_stid', 'display_on_screen', 'description_order', 'enabled', 'looping', 'header',
            'mute_objectives', 'conditions', 'effects'
        ]
        trigger_attr = {}
        for key in keys:
            if locals()[key] is not None:
                trigger_attr[key] = locals()[key]
        new_trigger = Trigger(name=name, trigger_id=len(self.triggers), **trigger_attr, host_uuid=self._host_uuid)
        self.triggers.append(new_trigger)
        return new_trigger

    def import_triggers(self, triggers: List[Trigger], index: int = -1, deepcopy: bool = True) -> List[Trigger]:
        """
        Adds existing trigger objects (from another scenario) to this scenario. Keeping all ``(de)activate trigger``
        effects linked!

        Args:
            triggers: The list of Trigger objects to be added
            index: The index where to insert the new triggers, will be added at the end when left unused.
            deepcopy: If the given triggers need to be deep copied or not when importing. Can be useful to keep the
                reference alive between the source and target trigger the same when setting this to `False`.

        Returns:
            The newly added triggers (with the new IDs and activation links etc.)
        """
        if deepcopy:
            triggers = copy.deepcopy(triggers)
        index_changes = {}

        for offset, trigger in enumerate(triggers):
            new_index = len(self.triggers) + offset
            index_changes[trigger.trigger_id] = trigger.trigger_id = new_index

        for trigger in triggers:
            for i, effect in enumerate(get_activation_effects(trigger)):
                try:
                    effect.trigger_id = index_changes[effect.trigger_id]
                except KeyError:
                    warn(f"(De)Activation effect {i} in trigger '{trigger.name}' refers to a trigger that wasn't "
                         f"included in the imported triggers. Effect will be reset")
                    effect.trigger_id = -1

        self.triggers += triggers
        if index != -1:
            self.move_triggers([t.trigger_id for t in triggers], index)
        return triggers

    def get_trigger(self, trigger_select: Union[int, TriggerSelect]) -> Trigger:
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)
        return trigger

    def remove_trigger(self, trigger_select: Union[int, TriggerSelect]) -> None:
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        del self.triggers[trigger_index]

        self.reorder_triggers()

    def _find_trigger_tree_nodes_recursively(self, trigger, known_node_indexes: List[int]) -> None:
        found_node_indexes = TriggerManager._find_trigger_tree_nodes(trigger)
        unknown_node_indexes = [i for i in found_node_indexes if i not in known_node_indexes]

        if len(unknown_node_indexes) == 0:
            return

        known_node_indexes += unknown_node_indexes

        for index in unknown_node_indexes:
            self._find_trigger_tree_nodes_recursively(self.triggers[index], known_node_indexes)

    def _validate_and_retrieve_trigger_info(self, trigger_select) -> (int, int, Trigger):
        if type(trigger_select) is int:
            trigger = display_index = None
            trigger_index = trigger_select
        else:
            trigger = trigger_select.trigger
            trigger_index = trigger_select.trigger_index
            display_index = trigger_select.display_index

        try:
            if trigger is not None:
                trigger_index = trigger.trigger_id
                display_index = self.trigger_display_order.index(trigger_index)
            elif trigger_index is not None:
                trigger = self.triggers[trigger_index]
                display_index = self.trigger_display_order.index(trigger_index)
            elif display_index is not None:
                trigger_index = self.trigger_display_order[display_index]
                trigger = self.triggers[trigger_index]
        except IndexError:
            if trigger_index:
                raise ValueError(f"No trigger with index {trigger_index}") from None
            if display_index:
                raise ValueError(f"No Trigger with display index {display_index}") from None

        return trigger_index, display_index, trigger

    def get_summary_as_string(self) -> str:
        return_string = "\nTrigger Summary:\n"

        triggers = self.triggers
        display_order = self.trigger_display_order

        if len(display_order) == 0:
            return_string += "\t<< No Triggers >>"

        longest_trigger_name = -1
        longest_index_notation = -1
        for display, trigger_index in enumerate(display_order):
            trigger_name = triggers[trigger_index].name
            longest_trigger_name = max(longest_trigger_name, len(trigger_name))

            longest_index_notation = max(
                longest_index_notation,
                helper.get_int_len(display) + helper.get_int_len(trigger_index)
            )

        longest_trigger_name += 3
        for display, trigger_index in enumerate(display_order):
            trigger = triggers[trigger_index]
            trigger_name = trigger.name

            name_buffer = longest_trigger_name - len(trigger_name)
            index_buffer = longest_index_notation - (helper.get_int_len(display) + helper.get_int_len(trigger_index))
            return_string += "\t" + trigger_name + (" " * name_buffer)
            return_string += f" [Index: {trigger_index}, Display: {display}] {' ' * index_buffer}"

            return_string += "\t(conditions: " + str(len(trigger.conditions)) + ", "
            return_string += " effects: " + str(len(trigger.effects)) + ")\n"

        return return_string

    def get_content_as_string(self) -> str:
        return_string = "\nTriggers:\n"

        if len(self.triggers) == 0:
            return_string += "\t<<No triggers>>\n"

        for trigger_index in self.trigger_display_order:
            return_string += self.get_trigger_as_string(trigger_index) + "\n"

        return return_string

    def get_trigger_as_string(self, trigger_select: Union[int, TriggerSelect]) -> str:
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        return_string = "\t'" + trigger.name + "'"
        return_string += " [Index: " + str(trigger_index) + ", Display: " + str(display_index) + "]" + ":\n"

        return_string += add_tabs(trigger.get_content_as_string(include_trigger_definition=False), 2)

        return return_string

    @staticmethod
    def _find_alterable_ce(trigger, trigger_ce_lock) -> (List[int], List[int]):
        lock_conditions = trigger_ce_lock.lock_conditions if trigger_ce_lock is not None else False
        lock_effects = trigger_ce_lock.lock_effects if trigger_ce_lock is not None else False
        lock_condition_type = trigger_ce_lock.lock_condition_type if trigger_ce_lock is not None else []
        lock_effect_type = trigger_ce_lock.lock_effect_type if trigger_ce_lock is not None else []
        lock_condition_ids = trigger_ce_lock.lock_condition_ids if trigger_ce_lock is not None else []
        lock_effect_ids = trigger_ce_lock.lock_effect_ids if trigger_ce_lock is not None else []

        alter_conditions: List[int] = []
        alter_effects: List[int] = []
        if not lock_conditions:
            for i, cond in enumerate(trigger.conditions):
                if i not in lock_condition_ids and cond.condition_type not in lock_condition_type:
                    alter_conditions.append(i)
        if not lock_effects:
            for i, effect in enumerate(trigger.effects):
                if i not in lock_effect_ids and effect.effect_type not in lock_effect_type:
                    alter_effects.append(i)

        return alter_conditions, alter_effects

    @staticmethod
    def _find_trigger_tree_nodes(trigger: Trigger) -> List[int]:
        return [
            effect.trigger_id for effect in trigger.effects if
            effect.effect_type in [EffectId.ACTIVATE_TRIGGER, EffectId.DEACTIVATE_TRIGGER]
        ]

    def __str__(self) -> str:
        return self.get_content_as_string()


def get_activation_effects(trigger: Trigger) -> List[Effect]:
    """
    Get all activation effects in a Trigger]

    Args:
        trigger (Trigger): The trigger object

    Returns:
        A list with (de)activation effects
    """
    return [eff for eff in trigger.effects if eff.effect_type in [
        EffectId.ACTIVATE_TRIGGER, EffectId.DEACTIVATE_TRIGGER
    ]]



import struct
from typing import Union, TYPE_CHECKING

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.helper.string_manipulations import has_str_trail, del_str_trail, q_str, \
    trunc_bytes, add_str_trail

if TYPE_CHECKING:
    from AoE2ScenarioParser.sections.retrievers.retriever import Retriever


def bytes_to_fixed_chars(byte_elements: bytes) -> str:
    end_of_string = byte_elements.find(b"\x00")
    if end_of_string != -1:
        byte_elements = byte_elements[:end_of_string]
    return bytes_to_str(byte_elements)


def fixed_chars_to_bytes(string: str, var_len: int) -> bytes:
    return str_to_bytes(string) + b"\x00" * (var_len - len(string))


# Update tests if this changes
# Tests *should* fail - but just to be sure
_no_string_trail = [
    "filename",                 # Scenario filename
    "ascii_instructions",       # Messages
    "ascii_hints",              # Messages
    "ascii_victory",            # Messages
    "ascii_loss",               # Messages
    "ascii_history",            # Messages
    "ascii_scouts",             # Messages
    "ascii_pregame",            # Cinematics
    "ascii_victory",            # Cinematics
    "ascii_loss",               # Cinematics
    "ascii_filename",           # Cinematics
    "strings",                  # in PlayerDataTwo
    "ai_names",                 # in PlayerDataTwo
    "ai_per_file_text",         # in AIStruct
    "unknown_string",           # in Map
    "map_color_mood",           # in Map
    "script_name",              # in Map
    "water_definition",         # in Map
    "message",                  # in Trigger  [ADDED SO THEY'RE IGNORED - HANDLED IN EFFECT COMMIT CALLBACK]
    "sound_name",               # in Trigger  [ADDED SO THEY'RE IGNORED - HANDLED IN EFFECT COMMIT CALLBACK]
    "script_file_path",         # in Files
    "script_file_content",      # in Files
]


def bytes_to_str(byte_elements, charset=settings.MAIN_CHARSET, fallback_charset=settings.FALLBACK_CHARSET) \
        -> Union[str, bytes]:
    """
    Converts bytes to string based on given charset.

    Args:
        byte_elements (bytes): Bytes to be decoded to string.
        charset (str): Main charset used to decode the bytes. Defaults settings.MAIN_CHARSET.
        fallback_charset (str): Fallback charset used to decode the bytes when the main fails. Defaults settings.FALLBACK_CHARSET.

    Returns:
        The decoded string or the byte elements when the string cannot be decoded
    """
    trailless_elements = del_str_trail(byte_elements) if has_str_trail(byte_elements) else byte_elements

    for c in [charset, fallback_charset]:
        try:
            return trailless_elements.decode(c)
        except ValueError:
            continue

    # Return the string as bytes when it cannot be decoded.
    # This will leave the string as-is.
    warn(f"Unable to decode bytes using '{charset}' and '{fallback_charset}', bytes: \n\t{trunc_bytes(byte_elements, 25)}")
    return byte_elements


def str_to_bytes(
        string: str,
        charset=settings.MAIN_CHARSET,
        fallback_charset=settings.FALLBACK_CHARSET
):
    """
    Converts string to bytes based on given charsets

    Args:
        string: The string to convert
        charset: The main charset used while encoding
        fallback_charset: The fallback charset when the main fails

    Returns:
        The converted string

    Raises:
        ValueError: When the string cannot be decoded with either of the charsets
    """
    for c in [charset, fallback_charset]:
        try:
            return string.encode(c)
        except ValueError:
            continue
    raise ValueError(f"Unable to encode string using '{charset}' and '{fallback_charset}'. String:\n\t{q_str(string)}")


def bytes_to_int(byte_elements, endian="little", signed=False):
    return int.from_bytes(byte_elements, endian, signed=signed)


def int_to_bytes(integer: int, length, endian="little", signed=True):
    return integer.to_bytes(length, byteorder=endian, signed=signed)


def bytes_to_float(byte_elements):
    return struct.unpack('f', byte_elements)[0]


def float_to_bytes(f):
    return struct.pack('f', f)


def bytes_to_double(byte_elements):
    return struct.unpack('d', byte_elements)[0]


def double_to_bytes(d):
    return struct.pack('d', d)


def parse_val_to_bytes(retriever: 'Retriever', val):
    var_type, var_len = retriever.datatype.type_and_length

    if var_type == "u" or var_type == "s":  # int
        return int_to_bytes(val, var_len, signed=(var_type == "s"))
    elif var_type == "str":  # str
        try:
            byte_string = str_to_bytes(val)
        except AttributeError as e:
            if type(val) is bytes:
                byte_string = val
            else:
                raise e from None

        if retriever.name not in _no_string_trail:
            byte_string = add_str_trail(byte_string)

        return _combine_int_str(byte_string, var_len, endian="little", signed=True, retriever=retriever)
    elif var_type == "c":  # str
        if len(val) > var_len:
            raise ValueError(f"Value cannot be longer than {var_len}, retriever: {retriever}")
        return fixed_chars_to_bytes(val, var_len)
    elif var_type == "data":  # bytes
        return val
    elif var_type == "f":  # float
        if var_len == 4:
            return float_to_bytes(val)
        else:
            return double_to_bytes(val)
    else:
        raise ValueError(f"Unable to parse value to bytes with unknown type: ({var_type})")


def parse_bytes_to_val(retriever, byte_elements):
    var_type, var_len = retriever.datatype.type_and_length

    if var_type == "u" or var_type == "s":
        return bytes_to_int(byte_elements, signed=(var_type == "s"))
    elif var_type == "str":
        return bytes_to_str(byte_elements[var_len:])
    elif var_type == "c":
        return bytes_to_fixed_chars(byte_elements)
    elif var_type == "data":
        return byte_elements
    elif var_type == "f":
        if var_len == 4:  # Float value
            return bytes_to_float(byte_elements)
        else:  # only 'else' is the trigger version. Which is a double (8 bytes)
            return bytes_to_double(byte_elements)
    else:
        raise ValueError(f"Unable to parse bytes with unknown type: ({var_type})")


def _combine_int_str(byte_string, length, endian, signed, retriever: 'Retriever' = None) -> bytes:
    try:
        return int_to_bytes(len(byte_string), length, endian=endian, signed=signed) + byte_string
    except OverflowError as e:
        if str(e) == "int too big to convert":
            name: str = retriever.name if retriever else ""
            raise OverflowError(
                f"{name.capitalize()} is too long. Length must be below {pow(2, length * 8 - 1)}, but is: {len(byte_string)}"
            ) from None



from typing import List, overload, Union, TYPE_CHECKING, Optional
from uuid import UUID

from AoE2ScenarioParser.datasets.players import PlayerId
from AoE2ScenarioParser.scenarios.scenario_store import store

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.data_objects.unit import Unit
    from AoE2ScenarioParser.objects.support.area import Area
    from AoE2ScenarioParser.objects.data_objects.trigger import Trigger


@overload
def unit_change_ownership(uuid: UUID, player: Union[int, PlayerId], unit: 'Unit') -> None:
    ...


@overload
def unit_change_ownership(uuid: UUID, player: Union[int, PlayerId], units: List['Unit']) -> None:
    ...


def unit_change_ownership(uuid: UUID, player: Union[int, PlayerId], *args) -> None:
    """
    Change the unit(s) ownership.

    Args:
        uuid: The UUID of the scenario
        player: The player to transfer the units to.
        args: Unit object or List of unit objects
    """

    def transfer_unit(scenario_, unit_, player_):
        scenario_.unit_manager.units[unit_.player].remove(unit_)
        scenario_.unit_manager.units[player_].append(unit_)

    scenario = store.get_scenario(uuid)
    if scenario:
        for units in args:
            if isinstance(units, List):
                for unit in units:
                    transfer_unit(scenario, unit, player)
            else:
                transfer_unit(scenario, units, player)


def import_triggers(uuid: UUID, triggers: List['Trigger'], insert_index: int = -1, deepcopy=True) -> None:
    """
    Import triggers into the scenario using the trigger manager.

    Args:
        uuid: The UUID of the scenario
        triggers: The trigger to import.
        insert_index: The insert index used in import triggers function
        deepcopy: If the given triggers need to be deep copied or not when importing. Can be useful to keep the
            reference alive between the source and target trigger the same when setting this to `False`.
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        scenario.trigger_manager.import_triggers(triggers, insert_index, deepcopy)


def remove_triggers(uuid: UUID, trigger_ids: List[int]) -> None:
    """
    Import triggers into the scenario using the trigger manager.

    Args:
        uuid: The UUID of the scenario
        trigger_ids: A list of trigger ids to remove.
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        scenario.trigger_manager.triggers = [
            trigger for trigger in scenario.trigger_manager.triggers if trigger.trigger_id not in trigger_ids
        ]
        scenario.trigger_manager.reorder_triggers()


def new_area_object(uuid: UUID) -> Optional['Area']:
    """
    Creates a new area object.

    Args:
        uuid: The UUID of the scenario
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.new.area()



from AoE2ScenarioParser.sections.dependencies.dependency_action import DependencyAction
from AoE2ScenarioParser.sections.dependencies.dependency_eval import DependencyEval
from AoE2ScenarioParser.sections.dependencies.dependency_target import DependencyTarget


class RetrieverDependency:
    def __init__(self, dependency_action, dependency_target=None, dependency_eval=None):
        """
        Object for handling dependencies between retrievers

        Args:
            dependency_action (DependencyAction): The type of action taken.
            dependency_target (DependencyTarget): The target of the action.
            dependency_eval (DependencyEval): Eval code to execute for action.
        """
        if dependency_action != DependencyAction.REFRESH_SELF and dependency_target is None:
            raise ValueError(f"Parameter dependency_target cannot be None when action is {dependency_action.name}")

        self.dependency_action = dependency_action
        self.dependency_target = dependency_target

        if dependency_action in [DependencyAction.SET_REPEAT, DependencyAction.SET_VALUE] and dependency_eval is None:
            # Get name of first (and only) target, must define eval if using more than one target
            dependency_eval = DependencyEval(dependency_target.targets[0][1])

        self.dependency_eval = dependency_eval

    @classmethod
    def from_structure(cls, structure):
        return cls(
            dependency_action=DependencyAction[structure.get('action')],
            dependency_target=DependencyTarget.instance_or_none(structure.get('target')),
            dependency_eval=DependencyEval.instance_or_none(structure.get('eval')),
        )

    def __repr__(self):
        repr_string = f"\n[RetrieverDependency] {self.dependency_action}"
        if self.dependency_target:
            repr_string += f"\n\tTarget: {self.dependency_target}"
        if self.dependency_eval:
            repr_string += f"\n\tEval: {self.dependency_eval.eval_code}"
        return repr_string



from pathlib import Path
from typing import Optional

from AoE2ScenarioParser.helper.exceptions import UnsupportedAttributeError, UnsupportedVersionError
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.trigger import Trigger
from AoE2ScenarioParser.scenarios.scenario_store import actions
from AoE2ScenarioParser.scenarios.scenario_store.getters import get_scenario_version
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.support import Support


class XsManagerDE(AoE2Object):
    """Manager of the everything XS related."""

    _link_list = [
        RetrieverObjectLink("script_name", "Map", "script_name", Support(since=1.40)),
    ]

    def __init__(self, script_name: str, **kwargs):
        super().__init__(**kwargs)

        self._script_name = script_name

        # --- XS Script Call Trigger ---
        self._initialized = False
        self._xs_trigger: Optional[Trigger] = Trigger(
            name="XS SCRIPT",
            enabled=False,
            description="Due to the lack of support for transferring XS files between systems in Age of Empires II:DE, "
                        "this trigger adds the entire script to an effect script call. This will add the script to"
                        "each system once the game starts in the default0.xs file. -- Created using AoE2ScenarioParser",
        )

    @property
    def script_name(self):
        return self._script_name

    @script_name.setter
    def script_name(self, value):
        if value:
            warn("Using script files added through `xs_manager.script_name` will not work in multiplayer.\n"
                 "You can work around this issue by using: `xs_manager.add_script(xs_file_path='path/to/script.xs')`.\n"
                 "For more information check out: https://ksneijders.github.io/AoE2ScenarioParser/cheatsheets/xs/")
        self._script_name = value

    @property
    def xs_trigger(self):
        if not self._initialized:
            self.initialise_xs_trigger()
        return self._xs_trigger

    @xs_trigger.setter
    def xs_trigger(self, value):
        self._xs_trigger = value

    def initialise_xs_trigger(self, insert_index: int = -1) -> None:
        """
        Creates the XS trigger on a desired location. If you don't care about the location, the `add_script()` function
        adds the trigger when calling it the first time too.

        If you want the trigger to be (almost) at the top of the list and you're reading a scenario with barely any to
        no triggers, it is recommended to call this somewhere at the start of the script.

        Insert index is used to move this trigger to a desired index.
        Keep in mind that moving triggers like this might take some time when you have a lot of triggers (thousands).

        Args:
            insert_index: The index where the xs trigger is added. Will be added at the end of the list if left as -1
        """
        try:
            self._xs_trigger.new_effect.script_call(message="")
        except UnsupportedAttributeError:
            raise UnsupportedVersionError(
                f"The scenario version ({get_scenario_version(self._host_uuid)}) does not support XS. "
                f"Save the scenario in the editor to update the scenario to allow for XS."
            ) from None
        self._initialized = True
        actions.import_triggers(self._host_uuid, [self.xs_trigger], insert_index, deepcopy=False)

    def _append_to_xs(self, title, string) -> None:
        self.xs_trigger.effects[0].message += f"// {'-' * 25} {title} {'-' * 25}\n{string}\n\n"

    def add_script(self, xs_file_path="", xs_string=""):
        """
        Add a script to the script call effect in the XS trigger

        Args:
            xs_file_path: Path to an XS file
            xs_string: Raw XS
        """
        if xs_file_path:
            path = Path(xs_file_path)
            with path.open() as xs_file:
                self._append_to_xs(path.name, xs_file.read())
        if xs_string:
            self._append_to_xs(f"XS string", xs_string)

    def _debug_write_script_to_file(self, filename="xs.txt"):
        with open(filename, 'w') as file:
            file.write(self.xs_trigger.effects[0].message)



from AoE2ScenarioParser.datasets.support.info_dataset_base import InfoDatasetBase


class BuildingInfo(InfoDatasetBase):
    """
    **Description**

    This is an enum class which provides information about most of the buildings in the game. Information about the
    following properties of a building is found in this class:
     - Unit ID
     - Icon ID
     - Dead Unit ID
     - HotKey ID
     - If the building is a gaia only unit (eg. ruins)

    **Inherited Methods from class InfoDatasetBase**

    >>> InfoDatasetBase.from_id()
    >>> InfoDatasetBase.from_dead_id()
    >>> InfoDatasetBase.from_icon_id()
    >>> InfoDatasetBase.from_hotkey_id()
    >>> InfoDatasetBase.gaia_only()
    >>> InfoDatasetBase.non_gaia()

    **Examples**

    >>> BuildingInfo.RUINS.ID
    >>> 345

    >>> BuildingInfo.ARCHERY_RANGE.ICON_ID
    >>> 0

    >>> BuildingInfo.BARRACKS.DEAD_ID
    >>> 1402

    >>> BuildingInfo.BLACKSMITH.HOTKEY_ID
    >>> 16131

    >>> BuildingInfo.RUINS.IS_GAIA_ONLY
    >>> True
    """
    RUINS = 345, -1, -1, 16393, True
    AACHEN_CATHEDRAL = 1622, 37, 1517, 16182, False
    AMPHITHEATRE = 251, 58, 1514, 16572, False
    AQUEDUCT = 231, 52, 1522, 16201, False
    ARCH_OF_CONSTANTINE = 899, 37, 1485, 16572, False
    ARCHERY_RANGE = 87, 0, 1415, 16128, False
    ARMY_TENT_A = 1196, 76, 1467, 16521, False
    ARMY_TENT_B = 1197, 76, 1468, 16521, False
    ARMY_TENT_C = 1198, 77, 1469, 16522, False
    ARMY_TENT_D = 1199, 77, 1470, 16523, False
    ARMY_TENT_E = 1200, 77, 1471, 16522, False
    BARRACKS = 12, 2, 1402, 16135, False
    BLACKSMITH = 103, 4, 1419, 16131, False
    BOMBARD_TOWER = 236, 42, 1439, 16156, False
    BRIDGE_A_BOTTOM = 607, -1, 144, 16513, False
    BRIDGE_A_BROKEN_BOTTOM = 740, -1, 144, 16643, False
    BRIDGE_A_BROKEN_TOP = 739, -1, 144, 16642, False
    BRIDGE_A_CRACKED = 738, -1, 144, 16641, False
    BRIDGE_A_MIDDLE = 606, -1, 144, 16512, False
    BRIDGE_A_TOP = 605, -1, 144, 16511, False
    BRIDGE_B_BOTTOM = 610, -1, 144, 16516, False
    BRIDGE_B_BROKEN_BOTTOM = 743, -1, 144, 16646, False
    BRIDGE_B_BROKEN_TOP = 742, -1, 144, 16645, False
    BRIDGE_B_CRACKED = 741, -1, 144, 16644, False
    BRIDGE_B_MIDDLE = 609, -1, 144, 16515, False
    BRIDGE_B_TOP = 608, -1, 144, 16514, False
    BRIDGE_C_BOTTOM = 1206, -1, 144, 16513, False
    BRIDGE_C_BROKEN_BOTTOM = 1212, -1, 144, 16643, False
    BRIDGE_C_BROKEN_TOP = 1211, -1, 144, 16642, False
    BRIDGE_C_CRACKED = 1210, -1, 144, 16641, False
    BRIDGE_C_MIDDLE = 1205, -1, 144, 16512, False
    BRIDGE_C_TOP = 1204, -1, 144, 16511, False
    BRIDGE_D_BOTTOM = 1209, -1, 144, 16516, False
    BRIDGE_D_BROKEN_BOTTOM = 1215, -1, 144, 16646, False
    BRIDGE_D_BROKEN_TOP = 1214, -1, 144, 16645, False
    BRIDGE_D_CRACKED = 1213, -1, 144, 16644, False
    BRIDGE_D_MIDDLE = 1208, -1, 144, 16515, False
    BRIDGE_D_TOP = 1207, -1, 144, 16514, False
    BRIDGE_E_BOTTOM = 1552, -1, 144, 16513, False
    BRIDGE_E_BROKEN_BOTTOM = 1558, -1, 144, 16643, False
    BRIDGE_E_BROKEN_TOP = 1557, -1, 144, 16642, False
    BRIDGE_E_CRACKED = 1556, -1, 144, 16641, False
    BRIDGE_E_MIDDLE = 1551, -1, 144, 16512, False
    BRIDGE_E_TOP = 1550, -1, 144, 16511, False
    BRIDGE_F_BOTTOM = 1555, -1, 144, 16516, False
    BRIDGE_F_BROKEN_BOTTOM = 1561, -1, 144, 16646, False
    BRIDGE_F_BROKEN_TOP = 1560, -1, 144, 16645, False
    BRIDGE_F_CRACKED = 1559, -1, 144, 16644, False
    BRIDGE_F_MIDDLE = 1554, -1, 144, 16515, False
    BRIDGE_F_TOP = 1553, -1, 144, 16514, False
    CASTLE = 82, 7, 1430, 16142, False
    CATHEDRAL = 599, 11, 1480, 16505, False
    CHAIN_WEST_TO_EAST = 1398, 72, 144, 16186, False
    CHAIN_SOUTHWEST_TO_NORTHEAST = 1396, 72, 144, 16186, False
    CHAIN_NORTH_TO_SOUTH = 1399, 72, 144, 16186, False
    CHAIN_NORTHWEST_TO_SOUTHEAST = 1397, 72, 144, 16186, False
    CITY_GATE_WEST_TO_EAST = 1587, 36, 1512, 16044, False
    CITY_GATE_SOUTHWEST_TO_NORTHEAST = 1579, 36, 1510, 16044, False
    CITY_GATE_NORTH_TO_SOUTH = 1591, 36, 1513, 16044, False
    CITY_GATE_NORTHWEST_TO_SOUTHEAST = 1583, 36, 1511, 16044, False
    CITY_WALL = 370, 31, 143, 16201, False
    COLOSSEUM = 263, 58, 1520, 16505, False
    DOCK = 45, 13, -1, 16144, False
    DORMITION_CATHEDRAL = 1369, 37, 1493, 16182, False
    FARM = 50, 35, 357, 16149, False
    FEITORIA = 1021, 53, 1446, 16159, False
    FENCE = 1062, 30, 1065, 16797, False
    FIRE_TOWER = 190, 26, 1438, 16158, False
    FISH_TRAP = 199, 41, 278, 16495, False
    FORTIFIED_PALISADE_WALL = 119, 30, 143, 16202, False
    FORTIFIED_TOWER = 1102, 45, 1444, 16231, False
    FORTIFIED_WALL = 155, 31, 1509, 16204, False
    FORTRESS = 33, 8, 1486, 16142, False
    GATE_NORTHWEST_TO_SOUTHEAST = 88, 36, 1501, 16185, False
    GATE_WEST_TO_EAST = 659, 36, 1502, 16185, False
    GATE_SOUTHWEST_TO_NORTHEAST = 64, 36, 1500, 16185, False
    GATE_NORTH_TO_SOUTH = 667, 36, 1503, 16185, False
    GOL_GUMBAZ = 1217, 37, 1487, 16578, False
    GUARD_TOWER = 234, 25, 1437, 16154, False
    HARBOR = 1189, 56, -1, 16144, False
    HOUSE = 70, 34, 1403, 16344, False
    HUT_A = 1082, 75, 1455, 16221, False
    HUT_B = 1083, 75, 1456, 16221, False
    HUT_C = 1084, 74, 1457, 16221, False
    HUT_D = 1085, 75, 1458, 16221, False
    HUT_E = 1086, 75, 1459, 16221, False
    HUT_F = 1087, 75, 1460, 16221, False
    HUT_G = 1088, 75, 1461, 16221, False
    KEEP = 235, 26, 1438, 16155, False
    KREPOST = 1251, 55, 1479, 16349, False
    LUMBER_CAMP = 562, 40, 1409, 16464, False
    MARKET = 84, 16, 1422, 16161, False
    MILL = 68, 19, 1404, 16157, False
    MINING_CAMP = 584, 39, 1410, 16487, False
    MONASTERY = 104, 10, 1421, 16138, False
    MONUMENT = 826, 37, -1, 16726, False
    OUTPOST = 598, 38, 1405, 16504, False
    PALISADE_GATE_SOUTHWEST_TO_NORTHEAST = 793, 44, 1441, 16186, False
    PALISADE_GATE_WEST_TO_EAST = 797, 44, 1442, 16186, False
    PALISADE_GATE_NORTHWEST_TO_SOUTHEAST = 789, 44, 1440, 16186, False
    PALISADE_GATE_NORTH_TO_SOUTH = 801, 44, 1443, 16186, False
    PALISADE_WALL = 72, 30, 1407, 16202, False
    PYRAMID = 689, 57, 1515, 16571, False
    QUIMPER_CATHEDRAL = 872, 37, 1489, 16572, False
    RICE_FARM = 1187, 35, 1188, 16149, False
    ROCK_CHURCH = 1378, 341, -1, 16182, False
    SANCHI_STUPA = 1216, 37, 1490, 16578, False
    SANKORE_MADRASAH = 1367, 37, 1491, 16182, False
    SEA_GATE_SOUTHWEST_TO_NORTHEAST = 1379, 71, -1, 16186, False
    SEA_GATE_NORTH_TO_SOUTH = 1391, 71, -1, 16186, False
    SEA_GATE_WEST_TO_EAST = 1387, 71, -1, 16186, False
    SEA_GATE_NORTHWEST_TO_SOUTHEAST = 1383, 71, -1, 16186, False
    SEA_TOWER = 785, 25, -1, 16704, False
    SEA_WALL = 788, 30, -1, 16707, False
    SHRINE = 1264, 12, 1483, 16138, False
    SIEGE_WORKSHOP = 49, 22, 1425, 16169, False
    STABLE = 101, 23, 1417, 16171, False
    STONE_WALL = 117, 31, 1508, 16203, False
    STORAGE = 1081, 59, 1484, 16219, False
    TEMPLE_OF_HEAVEN = 637, 11, 1481, 16505, False
    TENT_A = 1097, 78, 1462, 16230, False
    TENT_B = 1098, 83, 1463, 16230, False
    TENT_C = 1099, 83, 1464, 16230, False
    TENT_D = 1100, 83, 1465, 16230, False
    TENT_E = 1101, 83, 1466, 16230, False
    TOWER_OF_LONDON = 1368, 37, 1492, 16182, False
    TOWN_CENTER = 109, 28, 1408, 16164, False
    TRADE_WORKSHOP = 110, 17, 1429, 16174, False
    UNIVERSITY = 209, 32, 1427, 16176, False
    WATCH_TOWER = 79, 25, 1436, 16178, False
    WONDER = 276, 37, 1445, 16182, False
    WOODEN_BRIDGE_A_BOTTOM = 1311, -1, 144, 16513, False
    WOODEN_BRIDGE_A_MIDDLE = 1310, -1, 144, 16512, False
    WOODEN_BRIDGE_A_TOP = 1309, -1, 144, 16511, False
    WOODEN_BRIDGE_B_BOTTOM = 1314, -1, 144, 16516, False
    WOODEN_BRIDGE_B_MIDDLE = 1313, -1, 144, 16515, False
    WOODEN_BRIDGE_B_TOP = 1312, -1, 144, 16514, False
    YURT_A = 712, 81, 1447, 16594, False
    YURT_B = 713, 82, 1448, 16595, False
    YURT_C = 714, 82, 1449, 16596, False
    YURT_D = 715, 82, 1450, 16597, False
    YURT_E = 716, 80, 1451, 16598, False
    YURT_F = 717, 80, 1452, 16599, False
    YURT_G = 718, 80, 1453, 16600, False
    YURT_H = 719, 79, 1454, 16601, False
    DONJON = 1665, 84, 1524, 16350, False
    TRADE_WORKSHOP_BR = 1647, 17, 1429, 16174, False
    TRADE_WORKSHOP_TG = 179, 17, 1429, 16174, False
    TOWN_CENTER_PACKED = 444, 264, 1274, 16311, False
    FOLWARK = 1734, 86, 1528, 16157, False
    FORTIFIED_GATE_SOUTHWEST_TO_NORTHEAST = 63, 36, 1504, 16185, False
    FORTIFIED_GATE_NORTHWEST_TO_SOUTHEAST = 85, 36, 1505, 16185, False
    FORTIFIED_GATE_WEST_TO_EAST = 660, 36, 1506, 16185, False
    FORTIFIED_GATE_NORTH_TO_SOUTH = 668, 36, 1507, 16185, False
    PAGAN_SHRINE = 1712, 85, 1526, 16159, False
    GATE = 487, 36, 1406, 16185, False
    PALISADE_GATE = 792, 44, 1406, 16186, False
    FORTIFIED_GATE = 488, 36, 1406, 16185, False
    CARAVANSERAI = 1754, 87, 1529, 16351, False
    MINARET_OF_JAM = 1773, 37, 1530, 16182, False
